
//
//  aseRDB.cpp - Relational DataBase
//
//  This document contains file that you may use at your own risk in your own programs.
//
//  This module is designed to work with the exercises for this book.
//
//      This is the ASE Database
//
//  NOTE: Even though MAX uses XYZ, this uses XYZW and sets W to 0 to keep 128bit aligned!
//      Also, heavily overloaded with Assertions so any future expansion that breaks the
//      recursion and pointer validity, will be identified quickly.
//
//  Created - 16 Mar, 2002 - James Leiterman
//
//  Last Modification
//
//  stab=4
//

#include    "KariType.h"                                // Master Definitions

    //  3rd Party Files

#include    <iostream.h>
#include    <assert.h>
#include    <stdio.h>
#include    <string.h>
#include    <stdlib.h>
#include    <math.h>

    //  Book Development Library Modules
    
#include    "util.h"                                    // General purpose utilities
#include    "ram.h"                                     // Memory Module
#include    "vmp3D.h"
#include    "ase.h"                                     // ASE file parser
#include    "aseEx.h"                                   // ASE file parser (private data)







#define SOA_BLK_SIZE    4096



//
//  Double-Precision Floating-Point
//

#define SOAD_ARY_MAX    8               // array of 8 in a structure (# of fields in an array)
#define SOAD_BLK_MAX    (SOA_BLK_SIZE/(8*4*sizeof(double))) // 16=4096/256 # structures in block

#define SOAD_MAX        (SOAD_ARY_MAX * SOAD_BLK_MAX)

    //  256 bytes

typedef struct {
    double x[SOAD_ARY_MAX];
    double y[SOAD_ARY_MAX];
    double z[SOAD_ARY_MAX];
    double w[SOAD_ARY_MAX];
} SoAD;
    // # of Vectors per SoAD block!
#define SoA_DVEC_MAX    (sizeof(SoAD)/sizeof(vmp3DQDVector))    // =256/(4*8)= SOAD_ARY_MAX



    //  4096 Byte Block

typedef struct SoADBlk_Type
{
    SoAD blk[SOAD_BLK_MAX];
} SoADBlk;



//
//  Single-Precision Floating-Point
//
//  4096 Byte / 32 = 128byte blocks
//

#define SOA_ARY_MAX 8               // array of 8 in a structure (# of fields in an array)
#define SOA_BLK_MAX (SOA_BLK_SIZE/(8*4*sizeof(float)))  // 32=4096/128  # structures in block

#define SOA_MAX     (SOA_ARY_MAX * SOA_BLK_MAX)


    //  128 bytes

typedef struct {
    float x[SOA_ARY_MAX];
    float y[SOA_ARY_MAX];
    float z[SOA_ARY_MAX];
    float w[SOA_ARY_MAX];
} SoA;

#define SoA_VEC_MAX (sizeof(SoA)/sizeof(vmp3DQVector))  // = SOAD_ARY_MAX

    //  4096 Byte Block

typedef struct SoABlk_Type
{
    SoA blk[SOA_BLK_MAX];       //  insert other data arrays
} SoABlk;



//
//  Get # of bytes needed to construct 'n' blocks of (Double-Precision) SoA
//
//  NOTE: 1 nAoSCnt = 1 vmp3DQDVector
//
static uint vmp_GetSoADSize( uint nAoSCnt )
{
    return (nAoSCnt + (SoA_DVEC_MAX-1)) * sizeof(SoAD) / SoA_DVEC_MAX;
}



//
//  Get # of bytes needed to construct 'n' blocks of (Single-Precision) SoA
//
static uint vmp_GetSoASize( uint nAoSCnt )
{
    return (nAoSCnt + (SoA_VEC_MAX-1)) * sizeof(SoA) / SoA_VEC_MAX;
}



//
//  Double-Precision AoS (Array of Structures) to SoA (Structure of Arrays) Conversion
//
//  X0 X1 X2 X3 X4 X5 X6 X7         [] ={X0Y0Z0W0, ... X7Y7Z7W7};
//  Y0 Y1 Y2 Y3 Y4 Y5 Y6 Y7
//  Z0 Z1 Z2 Z3 Z4 Z5 Z6 Z7
//  W0 W1 W2 W3 W4 W5 W6 W7

static void vmp_SoAD_AoS( SoAD *pDAry, const vmp3DQDVector *pSAry, uint nVecCnt )
{
    uint n;
    double *pD, *pS;

    ASSERT_PTR(pSAry);
    ASSERT_PTR(pDAry);
    ASSERT_ZERO(nVecCnt);

    pS = (double*)pSAry;
    pD = (double*)pDAry;

        // 8 or more?
    if ( nVecCnt > SoA_DVEC_MAX )
      {
        n       = nVecCnt >> 3;     //   / SoA_DVEC_MAX;
        nVecCnt = nVecCnt & 0x07;   //   % SoA_DVEC_MAX; (reminder)

            // One complete 8x {XYZW} structure each pass!
        do {    //      .X                .Y                  .Z                 .W
            *(pD+0)=*(pS+0);  *(pD+8)=*(pS+1);   *(pD+16)=*(pS+2);  *(pD+24)=*(pS+3);
            *(pD+1)=*(pS+4);  *(pD+9)=*(pS+5);   *(pD+17)=*(pS+6);  *(pD+25)=*(pS+7);
            *(pD+2)=*(pS+8);  *(pD+10)=*(pS+9);  *(pD+18)=*(pS+10); *(pD+26)=*(pS+11);
            *(pD+3)=*(pS+12); *(pD+11)=*(pS+13); *(pD+19)=*(pS+14); *(pD+27)=*(pS+15);
            *(pD+4)=*(pS+16); *(pD+12)=*(pS+17); *(pD+20)=*(pS+18); *(pD+28)=*(pS+19);
            *(pD+5)=*(pS+20); *(pD+13)=*(pS+21); *(pD+21)=*(pS+22); *(pD+29)=*(pS+23);
            *(pD+6)=*(pS+24); *(pD+14)=*(pS+25); *(pD+22)=*(pS+26); *(pD+30)=*(pS+27);
            *(pD+7)=*(pS+28); *(pD+15)=*(pS+29); *(pD+23)=*(pS+30); *(pD+31)=*(pS+31);
            pS += SOAD_BLK_MAX; // 32
            pD += SOAD_BLK_MAX;
          } while(--n);
      }

        //  Handle any remainder {XYZW} structures
        //  NOTE: pD has room for one complete SoAD block!

    if (nVecCnt)
      {
        ramZero(pD, sizeof(SoAD));                      // Clear one full block

       do {     //      .X                .Y                  .Z                 .W
            *(pD+0)=*(pS+0);  *(pD+8)=*(pS+1);   *(pD+16)=*(pS+2);  *(pD+24)=*(pS+3);
            pS += 4;
            pD++;
          } while(--nVecCnt);
      }
}



//
//  Node Dump
//
void aseDump( const Node * pNode )
{
    uint n, nTotShapeCnt, nTotShapeVertexCnt;
    NodeScene *pScene;
    NodeMaterial *pMaterial;
    SubMaterial *pSubMaterial;
    GeomObject *pGeomObject;
    Mesh    *pMesh;
    ShapeObject *pShapeObject;
    ShapeLine *pShapeLine;

    while(NULL != pNode)
      {
        printf( "%s : ", AseKeyTbl[pNode->keyId].pszName );

        if (NULL == pNode->pName)
          {
            printf( " NO NAME NODE type" );
          }
        else
          {
            printf( " %s", pNode->pName);
          }

        switch(pNode->keyId)
          {
        case iKEY_SCENE:
            pScene = (NodeScene*)(pNode->pData);
            
            printf( " FirstFrame#%u LastFrame#%u speed:%u Tic/Frm:%u\n", 
                pScene->nFirstFrame, pScene->nLastFrame, pScene->nFrameSpeed, pScene->TicksPerFrame );
            break;

        case iKEY_MATERIAL_LIST:
            pMaterial = (NodeMaterial*)(pNode->pData);
            pSubMaterial = pMaterial->pSubMaterial;

            printf( "#%u of materials: %s\n", pMaterial->nItemCnt, pMaterial->pName  );
            for (n=0; n<pMaterial->nItemCnt; n++, pSubMaterial++ )
              {
                printf( "%s, %s, %s\n", pSubMaterial->pName, pSubMaterial->pMapName, pSubMaterial->pFPath );
              }
            printf( "\n" );
            break;

        case iKEY_GEOMOBJECT:
            pGeomObject = (GeomObject *)(pNode->pData);

            printf( "%s ", pGeomObject->pName );

            pMesh = pGeomObject->pAni;
            for (n=0; n < pGeomObject->nItemCnt; n++)
              {
                printf( "NumVertex:%u NumFaces:%u NumTVertex:%u NumTFaces:%u NumCVertex:%u NumCFaces:%u NumNormals:%u\n",
                    pMesh->NumVertex, pMesh->NumFaces, pMesh->NumTVertex, pMesh->NumTFaces,
                    pMesh->NumCVertex, pMesh->NumCFaces, pMesh->NumNormals );
              }
            break;

        case iKEY_SHAPEOBJECT:
            pShapeObject = (ShapeObject *)(pNode->pData);

            pShapeLine = pShapeObject->pLine[0];

            for (n=0; n<pShapeObject->nLineCnt; n++)
              {
                printf( " %u ", pShapeLine->nVertexCnt ); 

                pShapeLine++;
              }
            printf("\n" );
            break;

        default:
            printf( "\n" );
            break;
          }


        pNode = pNode->next;
      }
}



//
//  Repattern Vertices in memory from AoS into SoA
//
static bool aseAoStoSoA( const Node * pNode, rdbInfo *pStat )
{
    uint n, nSize, nCnt;
    NodeScene *pScene;
    NodeMaterial *pMaterial;
    SubMaterial *pSubMaterial;
    GeomObject *pGeomObject;
    Mesh    *pMesh;
    ShapeObject *pShapeObject;
    ShapeLine *pShapeLine;
    SoAD *pSoA;

    printf( "Total Shapes=%u  Total ShapeVertices=%u\n", pStat->nTotShapeCnt, pStat->nTotShapeVertexCnt );

        //  If any vertices, then shapes, thus, Allocate Memory for Master Table for all shapes

    if ( pStat->nTotShapeVertexCnt )
      {
        if (!ramGetClr((byte**)&(pStat->pShapeLookup), pStat->nTotShapeCnt * sizeof(rdbPos)))
          {
            return false;
          }

        nCnt = pStat->nTotShapeVertexCnt;               // Align to SoA Block

        if (!ramGetClr((byte**)&(pStat->pVertShapeList), nCnt * sizeof(vmp3DQVector)))
          {
            return false;
          }
      }

        //  Now parse list

    while(NULL != pNode)
      {
        switch(pNode->keyId)
          {
            //  Ignore Scene Data This pass
        case iKEY_SCENE:
            break;

            //  Ignore Materials This pass
        case iKEY_MATERIAL_LIST:
            break;

        case iKEY_GEOMOBJECT:
            pGeomObject = (GeomObject *)(pNode->pData);

            pMesh = pGeomObject->pAni;
            for (n=0; n < pGeomObject->nItemCnt; n++)
              {
                    //  _____________________________
                    //  Process AoS to SoA 'Vertices'
                    //  Note that Vertice coordinates will be kept isolated from vertice and face properties!
                    //  Allocate new memory
                if (pMesh->NumVertex)
                  {
                    nSize = vmp_GetSoADSize( pMesh->NumVertex );
                    if (!ramGet((byte **)&pSoA, nSize))
                      {
                        return false;
                      }

                        //  Shuffle AoS 'Vertices' into SoA pattern
        
                    vmp_SoAD_AoS( pSoA, pMesh->pVertList, pMesh->NumVertex ); 

                    ramRelease((byte**)&(pMesh->pVertList));    // Destroy old list
                    pMesh->pVertList = (vmp3DQDVector*)pSoA;    // Assign new Double-Precision SoA data
                  }

                    //  Leave TVertices, CVertices, & Normals, in AoS 0.0 ... 1.0
              }
            break;

            //  Keep shape Lines (rails) intact

        case iKEY_SHAPEOBJECT:
            pShapeObject = (ShapeObject *)(pNode->pData);
            pShapeLine = pShapeObject->pLine[0];

            for (n=0; n<pShapeObject->nLineCnt; n++)
              {
                printf( " %u ", pShapeLine->nVertexCnt ); 

                pShapeLine++;
              }
            break;

        default:
            break;
          }

        pNode = pNode->next;
      }

    return true;
}


//
//  ASE Relational Database
//
//  Data Tree is in AoS (Array of Structures) Double-Precision Floating-Point
//  We almost need to manipulate the data, at the very least scale it for our
//  game environment so first we need to convert from AoS to SoA keeping data
//  in 64bit DP.
//
//  process the data but initially keep
//  in a Double-Precision Floating-Point format. 
//
//  Then convert from Double-Precision to Single-Precision.
//
//  Then write that data to a file.

bool aseRDB( const Node * pNode, rdbInfo *pStat )
{

aseDump(pNode);

        //  Repattern Vertices coordinates in memory from AoS into SoA

    if (!aseAoStoSoA( pNode, pStat ))
      {
        ELOG( "Error AoS to SoA Double-Precision Conversion\n" );
        return false;
      }






    return true;
}


